package main

import (
	"encoding/json"
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"os"
	"path/filepath"
	"strconv"

	"github.com/labstack/echo"
	"github.com/labstack/echo/middleware"
)

const (
	CMD_QUIT = "quit"
)

var (
	myserver_root = "./"
	debug_mode    = false
)

// 設定情報
type Config struct {
	Host        string           `json:"host"`
	Port        uint16           `json:"port"`
	Control_url string           `json:"control_url"`
	Actions     []ExecutorAction `json:"actions"`
}

// リクエストハンドラ：バージョン返却
func requestHandler(c echo.Context) error {
	return c.String(http.StatusOK, "IoT Automation Ver0.1")
}

// リクエストハンドラ：プログラムコントロール
func controlHandler(c echo.Context) error {
	if debug_mode {
		fmt.Println("control handler")
	}
	// コマンド名取得
	cmd := c.FormValue("command")
	// コマンド毎の処理
	switch cmd {
	case CMD_QUIT: // プログラムの終了
		if debug_mode {
			fmt.Println(CMD_QUIT)
		}
		// 自動処理を停止
		stopExecutor()
		// プログラム終了
		os.Exit(0)
	}
	return c.NoContent(http.StatusOK)
}

// メイン処理
func main() {

	// プログラム配置場所のTOPを取得
	myserver_root = os.Getenv("MYSERVER_ROOT")
	log.Println("MYSERVER_ROOT = ", myserver_root)

	/*** プログラム動作の設定ロード ***/
	// 設定ファイルを読み込むパスを設定
	prog_dir := "./"
	if len(myserver_root) == 0 {
		prog_dir = filepath.Dir(os.Args[0])
	} else {
		prog_dir = filepath.Join(myserver_root, "MyIoTService/Automate")
	}
	config_path := filepath.Join(prog_dir, "config.json")
	log.Println(config_path)
	// 設定ファイルを読み込む
	conf := new(Config)
	conf_file, err := ioutil.ReadFile(config_path)
	if err != nil {
		conf_file, err = ioutil.ReadFile("./config.json")
		if err != nil {
			log.Println("Config file read error. - ", err)
			os.Exit(1)
		}
	}
	err = json.Unmarshal(conf_file, conf)
	if debug_mode {
		fmt.Println(conf)
	}

	/*** 自動実行処理エンジンの起動 ***/
	// アクション実行Goルーチンの起動
	for no, act := range conf.Actions {
		fmt.Println(no, "execute action of ", act)
		go MQTTMessageProcessor(act)
	}

	/*** コントロール用のAPIサーバ初期化 ***/
	// Echoセットアップ
	e := echo.New()
	// エラー発生時の対処設定
	e.Use(middleware.Recover())

	// ブラウザからjavascriptを使ってAPI呼び出しできるようにCORS対応
	e.Use(middleware.CORSWithConfig(middleware.CORSConfig{
		AllowOrigins: []string{"*"},
		AllowMethods: []string{echo.GET, echo.PUT, echo.POST, echo.DELETE, echo.HEAD},
	}))

	// リクエストハンドラ登録
	e.GET("/", requestHandler)
	e.POST(conf.Control_url, controlHandler)

	// サーバの起動
	e.Logger.Fatal(e.Start(conf.Host + ":" + strconv.FormatInt(int64(conf.Port), 10)))

}
